#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <net-snmp/net-snmp-config.h>
#include <net-snmp/net-snmp-includes.h>
#include <opservices.h>

char *get_interface_index( struct snmp_session *snmp, char *if_name );
int do_snmp_request( struct snmp_session *snmp, char *oidToRead );
void parse_getopt( struct snmp_session *snmp, char **ifIndex, char **ifLabel, int argc, char **args );
void show_help();

int main( int argc, char **argv ) {

	int oStatus;
	int aStatus;
	int retCode = OK;
	char *ifIndex = NULL;
	char *ifName = NULL;
	char operStatus[] = ".1.3.6.1.2.1.2.2.1.8";
	char adminStatus[] = ".1.3.6.1.2.1.2.2.1.7";
	char *tmpOid;
	struct snmp_session session, *ss;

	// Create a SNMP session
	init_snmp("snmpapp");
	snmp_sess_init( &session );
	
	// Parse getopt and set SNMP session
	parse_getopt( &session, &ifIndex, &ifName, argc, argv );

	// At least one of these parameters must be provided
	if (ifName == NULL && ifIndex == NULL) {
		show_help();
	}

	// Start SNMP session
	ss = snmp_open( &session );
	if (!ss) {
		snmp_perror("ack");
		snmp_log(LOG_ERR, "Unable to initiate SNMP session\n");
		exit(CRITICAL);
	}

	if (ifName != NULL && ifIndex == NULL) 
		ifIndex = get_interface_index(ss,ifName);

	if (ifIndex == NULL) {
		printf("Unable to locate ifIndex for interface %s\n",ifName);
		exit(UNKNOWN);
	}

	// Get operStatus
	asprintf(&tmpOid, "%s.%s", operStatus, ifIndex);
	if ( (oStatus = do_snmp_request(ss,tmpOid)) == ERROR ) {
		printf("Unable to retrieve operStatus for interface\n");
		exit(UNKNOWN);
	}
	
	// Get adminStatus
	asprintf(&tmpOid, "%s.%s", adminStatus, ifIndex);
	if ( (aStatus = do_snmp_request(ss,tmpOid )) == ERROR ) {
		printf("Unable to retrieve adminStatus for interface\n");
		exit(UNKNOWN);
	}

	// Close SNMP session
	snmp_close(ss);

	// Output results
	if (ifName == NULL)
		printf("If_%s - ", ifIndex);
	else
		printf("If_%s - ", ifName);
		
	if (oStatus == 1) {
		printf("operStatus(up) - ");
	} else {
		printf("operStatus(down) - ");
		retCode = CRITICAL;
	}
		 
	if (aStatus == 1) { 
		printf("adminStatus(up)\n");
	} else { 
		printf("adminStatus(down)\n");
	}
	exit(retCode);

}

void parse_getopt( struct snmp_session *snmp, char **ifIndex, char **ifLabel, int argc, char **args ) {

	int c;
	int version;

	while ( ( c = getopt( argc, args, "H:c:i:l:v:h" )) != -1 ) {

		switch(c) {

		case 'H':
			snmp->peername = optarg;
			break;

		case 'c':
			snmp->community = (u_char *)optarg;
			snmp->community_len = strlen(optarg);
			break;

		case 'l':
			*ifLabel = optarg;
			break;

		case 'i':
			*ifIndex = optarg;
			break;

		case 'v':
			version = atoi(optarg);
			if (version == 1) {
				snmp->version = SNMP_VERSION_1;
			} else if (version == 2) {
				snmp->version = SNMP_VERSION_2c;
			}
			break;

		case '?':
		case 'h':
		default:
			show_help();
			break;

		}
	}

}

void show_help() {

	printf("check_snmp_ifstatus (opservices-plugins) 1.0\n");
	printf("Copyright (c) 2009 OpServices Development Team\n");
	printf("\t\t<desenv@opservices.com.br>\n\n");
	printf("Get interface's operStatus and adminStatus via SNMP.\n\n");
	printf("USAGE: check_snmp_ifstatus -H <ip_address> -c <community> -v <snmp_version>\n");
	printf("                           -i <interface_index> [ -l <interface_label> ]\n");
	printf("Options:\n");
	printf(" -H\n\tHost name or IP Address\n");
	printf(" -c\n\tCommunity string for SNMP communication\n");
	printf(" -v\n\tSNMP Version\n");
	printf(" -i\n\tInterface index\n");
	printf(" -l\n\tInterface alias to show a better output\n\n");
	exit(WARNING);

}

int do_snmp_request( struct snmp_session *snmp, char *oidToRead ) {

	int status;
	int result = ERROR;
	struct snmp_pdu *pdu;
	struct snmp_pdu *response;
	struct variable_list *vars;
	size_t anOidLen = MAX_OID_LEN;
	oid anOid[MAX_OID_LEN];
	
	// Set SNMP request settings
	pdu = snmp_pdu_create(SNMP_MSG_GET);
	read_objid(oidToRead, anOid, &anOidLen);
	snmp_add_null_var(pdu, anOid, anOidLen);

	// Send request and get response
	status = snmp_synch_response(snmp, pdu, &response);
	if (status == STAT_SUCCESS && response->errstat == SNMP_ERR_NOERROR) {	
		vars = response->variables;
		if (vars->type == ASN_INTEGER) {
			result = (int) *(vars->val.integer);
		}
	} else {
		if (status == STAT_SUCCESS)
			fprintf(stderr, "Error in packet\nReason: %s\n", snmp_errstring(response->errstat));
		else
			snmp_sess_perror("snmpget", snmp);
	}

	// Free resources
	snmp_free_pdu(response);

	return result;

}

char *get_interface_index( struct snmp_session *snmp, char *if_name ) {

	int status;
	int running = TRUE;
	struct snmp_pdu *pdu;
	struct snmp_pdu *response;
	struct variable_list *vars;
	size_t anOidLen = MAX_OID_LEN;
	oid anOid[MAX_OID_LEN];
	size_t rootLen = MAX_OID_LEN;
	oid root[MAX_OID_LEN];
	char *ret = NULL;

	// Set SNMP request settings
	if (snmp_parse_oid(".1.3.6.1.2.1.31.1.1.1.1", root, &rootLen) == NULL) {
		snmp_perror(".1.3.6.1.2.1.31.1.1.1.1");
		exit(WARNING);
	}

	memmove(anOid, root, rootLen * sizeof(oid));
	anOidLen = rootLen;

	while(running) {

		pdu = snmp_pdu_create(SNMP_MSG_GETNEXT); 
		snmp_add_null_var(pdu, anOid, anOidLen);

		// Send request and get response
		status = snmp_synch_response(snmp, pdu, &response);
		if (status == STAT_SUCCESS && response->errstat == SNMP_ERR_NOERROR) {	

			for ( vars = response->variables; vars; vars = vars->next_variable ) {

				if ((vars->name_length < rootLen) || (memcmp(root, vars->name, rootLen * sizeof(oid)) != 0)) {
					running = FALSE;
					continue;
				}

				if (vars->type == ASN_OCTET_STR) {

					char *str = (char *)malloc(1 + vars->val_len);
					memcpy(str, vars->val.string, vars->val_len);
					str[vars->val_len] = '\0';

					memmove((char *) anOid, (char *) vars->name, vars->name_length * sizeof(oid));
					anOidLen = vars->name_length;

					if (strcmp(if_name,str) == 0) {

						char *buf = (char *)malloc(anOidLen + 1);
						snprint_objid( buf, sizeof(anOid), anOid, anOidLen );
						char *result = NULL;
						result = strtok( buf, "." );
						if (result != NULL) {
							result = strtok( NULL, "." );	
							ret = result;
						}

					}
				}

			}
		} else {
			if (status == STAT_SUCCESS)
				fprintf(stderr, "Error in packet\nReason: %s\n", snmp_errstring(response->errstat));
			else
				snmp_sess_perror("snmpget", snmp);
			running = FALSE;
		}

	}

	// Free resources
	snmp_free_pdu(response);

	return ret;

}
