#include <opservices.h>

extern char			*op_snmp_error;
struct snmp_check_parameters 	*params = NULL;
static const char		 default_warning[] = "3,3,3";
static const char		 default_critical[] = "4,4,4";

struct cpu_metric {
	struct cpu_metric	*next;
	char			*name;
	double			 value;
	double			 warning;
	double			 critical;
};

int
op_snmp_parse_args(int argc, char **argv, struct snmp_check_parameters *params)
{

	int	arg;
	int	ret;

	ret = init_snmp_check_params(params);
	if (ret != OK)
		return ret;

	/* default values for warning and critical */
	params->warning_str = strdup(default_warning);
	params->critical_str = strdup(default_critical);

	while( (arg = getopt(argc,argv,"i:h:w:c:t:r:d")) != EOF )  {

		switch(arg) {

		case 'i':
			params->community_id = atoi(optarg);
			break;

		case 'h':
			params->host_ip = strdup(optarg);
			memset(optarg, '\x0', strlen(optarg));
			break;

		case 'w':
			params->warning_str = optarg;
			break;

		case 'c':
			params->critical_str = optarg;
			break;

		case 't':
			params->timeout = atoi(optarg);
			break;

		case 'r':
			params->retries = atoi(optarg);
			break;

		case 'd':
			DEBUG = TRUE;
			break;

		}

	}

	if (params->community_id == 0 || params->host_ip == NULL)
		return ERROR;

	return OK;

}

void
add_to_ll(struct cpu_metric **tree, struct cpu_metric *node)
{

	struct cpu_metric	*cursor;

	if (*tree == NULL) {
		*tree = node;
		return;
	}

	cursor = *tree;
	while(cursor->next) {
		cursor = cursor->next;
	}

	cursor->next = node;
}

int
usage() 
{
	printf("Usage:\n");
	printf("op-snmp-load-unix -h <host ip> -i <community id> [-w <warning>] [-c <critical>] [-t <snmp timeout>] [-r <retries>]\n");
	return UNKNOWN;
}

int
get_metrics(struct cpu_metric **tree, struct snmp_session *session)
{

	oid 			 metric_names[] = { 1, 3, 6, 1, 4, 1, 2021, 10, 1, 2 }; 
	struct cpu_metric	*m = NULL;
	struct snmp_leaf 	*root_leaf = NULL;
	struct snmp_leaf 	*leaf = NULL;

	root_leaf = op_snmp_walk(metric_names, sizeof(metric_names), session);
	if (root_leaf == NULL && op_snmp_error != NULL) {
		printf("%s\n",op_snmp_error);
		return UNKNOWN;
	}

	for(leaf = root_leaf; leaf; leaf = leaf->next) {
		m = calloc(1, sizeof(struct cpu_metric));
		m->name = strdup(leaf->_char);
		m->next = NULL;

		add_to_ll(tree, m);
	}

	if (DEBUG) {
		printf("Metrics found:\n");
		m = *tree;
		while(m) {
			printf("[%s]\n", m->name);
			m = m->next;
		}
	}

	if (*tree == NULL) {
		printf("UNKNOWN - Unable to locate load counters through snmp\n");
		return UNKNOWN;
	}

	return OK;

}

double
explode_threshold(char *values, int pos)
{

	char	*delim = NULL;
	char	*buf_ptr = NULL;

	buf_ptr = strdup(values);
	while(TRUE) {

		delim = strchr(buf_ptr,',');
		if (delim == NULL) {
			/* no more delimiters */
			if (DEBUG)
				printf("returning %.0f\n", atof(buf_ptr));
			return atof(buf_ptr);
		}

		*delim = '\x0';

		if (pos == 0) {
			if (DEBUG)
				printf("returning %.0f\n", atof(buf_ptr));
			return atof(buf_ptr);
		}

		buf_ptr = delim + 1;
		pos--;

	}

}

void
add_value_to_node(struct cpu_metric **tree, double value, int pos)
{
	struct cpu_metric	*m;
	int			 pos_bkp;

	pos_bkp = pos;
	m = *tree;
	while(pos > 0) {
		m = m->next;
		pos--;
	}

	m->value = value;
	m->warning = explode_threshold(params->warning_str, pos_bkp);
	m->critical = explode_threshold(params->critical_str, pos_bkp);

}

int
get_values(struct cpu_metric **tree, struct snmp_session *session)
{
	oid 	 		 metric_values[] = { 1, 3, 6, 1, 4, 1, 2021, 10, 1, 3 }; 
	int			 count = 0;
	double			 value;
	struct cpu_metric	*m = NULL;
	struct snmp_leaf 	*root_leaf = NULL;
	struct snmp_leaf 	*leaf = NULL;

	root_leaf = op_snmp_walk(metric_values, sizeof(metric_values), session);
	if (root_leaf == NULL && op_snmp_error != NULL) {
		printf("%s\n",op_snmp_error);
		return UNKNOWN;
	}

	for(leaf = root_leaf; leaf; leaf = leaf->next) {
		value = atof(leaf->_char);
		add_value_to_node(tree, value, count);
		count++;
	}

	if (DEBUG) {
		printf("Values found:\n");
		m = *tree;
		while(m) {
			printf("[%f]\n", m->value);
			m = m->next;
		}
	}

	return OK;

}

int
main(int argc, char **argv)
{

	struct snmp_session 		 session;
	int 				 status;
	char 				*plugin_output = NULL;
	char				*metric_output = NULL;
	char				*aux = NULL;
	struct cpu_metric		*root = NULL;
	struct cpu_metric		*m;

	op_get_memcache_servers();
	parse_db_conf();

	init_snmp("op-snmp-load-unix");

	params = calloc(1, sizeof(struct snmp_check_parameters));
	if (params == NULL) {
		printf("Unable to allocate memory to parse command line parameters\n");
		return UNKNOWN;
	}

	status = op_snmp_parse_args(argc,argv,params);
	if (status != OK)
		return usage();

	snmp_sess_init(&session);
	status = prepare_snmp_session(&session, params, NULL);
	if (status != OK) {
		printf("Unable to locate community id %d\n",params->community_id);
		return UNKNOWN;
	}

	status = get_metrics(&root, &session);
	if (status != OK)
		return status;

	status = get_values(&root, &session);
	if (status != OK)
		return status;

	status = OK;
	m = root;
	asprintf(&plugin_output,"Return:");
	asprintf(&metric_output,"|");
	while(m) {
		if (m->value > m->critical)
			status = CRITICAL;
		else if (m->value > m->warning && status != CRITICAL)
			status = WARNING;

		aux = plugin_output;
		asprintf(&plugin_output, "%s %s: %.2f", plugin_output, m->name, m->value);
		free(aux);

		aux = metric_output;
		asprintf(&metric_output, "%s%s=%.2f;%.2f;%.2f;; ", 
		    metric_output, 
		    m->name, 
		    m->value,
		    m->warning,
		    m->critical
		);
		free(aux);

		m = m->next;
	}
	
	printf("%s %s\n", plugin_output, metric_output);

	return status;

}
