#include <opservices.h>

extern char *op_snmp_error;

int op_snmp_parse_args(int argc, char **argv, struct snmp_check_parameters *params) {

	int arg;
	int ret;

	ret = init_snmp_check_params(params);
	if (ret != 0) {
		return ret;
	}

	while ( (arg = getopt(argc,argv,"i:h:w:c:f:t:r:d")) != EOF )  {

		switch(arg) {

		case 'i':
			params->community_id = atoi(optarg);
			break;

		case 'h':
			params->host_ip = strdup(optarg);
			memset(optarg, '\x0', sizeof(optarg));
			break;

		case 'w':
			params->warning = atof(optarg);
			break;

		case 'c':
			params->critical = atof(optarg);
			break;

		case 'd':
			DEBUG = TRUE;
			break;

		case 't':
			params->timeout = atoi(optarg);
			break;

		case 'r':
			params->retries = atoi(optarg);
			break;

		case 'f':
			params->filter = strdup(optarg);
			memset(optarg, '\x0', strlen(optarg));
			break;

		}

	}

	if ( params->community_id == 0 || 
	     params->host_ip == NULL ||
	     params->filter == NULL ) {

		return ERROR;
	}

	return OK;


}

int usage() {
	printf("Usage:\n");
	printf("op-snmp-storage-stats -h <host ip> -i <community id> -w <warning> -c <critical> -f <disk name> [-t <snmp timeout>] [-r <snmp retries>]\n");
	return UNKNOWN;
}


char *prepare_disk_size_output(float used, float size, float unit) {

	int size_unit    = 0;
	int used_unit    = 0;
	int i            = 0;
	int	diff         = 0;
	float used_bytes = 0;
	float size_bytes = 0;
	char *ret        = NULL;
	char *units[6]   = { "bytes",
	                      "KB",
	                      "MB",
	                      "GB",
	                      "TB",
	                      "EB"
	};

	used_bytes = used * unit;
	size_bytes = size * unit;

	while(size_bytes >= 1024) {
		size_unit++;
		size_bytes /= 1024;
	}

	if (size_unit > 5) {
		diff = size_unit - 5;
		size_unit = 5;
		for(i=0; i<diff; i++) {
			size_bytes *= 1024;
		}
	}

	while(used_bytes >= 1024) {
		used_unit++;
		used_bytes /= 1024;
	}
	if (used_unit > 5) {
		diff = used_unit - 5;
		used_unit = 5;
		for(i=0; i<diff; i++) {
			used_bytes *= 1024;
		}
	}

	asprintf(&ret,
	    "%.1f%s/%.1f%s",
	    used_bytes,
	    units[used_unit],
	    size_bytes,
	    units[size_unit]
	);

	return ret;

}

int main(int argc, char **argv) {

	struct snmp_session session;
	char plugin_output[MAX_PLUGIN_OUTPUT_LENGTH];
	float disk_size  = 0;
	float disk_usage = 0;
	float disk_unit  = 0;
	float usage_pct  = 0;
	float mem_buffer;
	float mem_cache;
	int status;
	struct snmp_check_parameters *params = NULL;
	struct snmp_leaf *root_leaf = NULL;
	struct snmp_leaf *leaf = NULL;
	unsigned long index = 0;
	regex_t disk_name_regex;
	regex_t system_operation_regex;
	char *disk_regex_string;
	bool isLinux = FALSE;
	oid disks_names_base_oid[] = { 1, 3, 6, 1, 2, 1, 25, 2, 3, 1, 3 };
	oid disk_max_size[]        = { 1, 3, 6, 1, 2, 1, 25, 2, 3, 1, 5, 0 };
	oid disk_used_size[]       = { 1, 3, 6, 1, 2, 1, 25, 2, 3, 1, 6, 0 };
	oid disk_size_unit[]       = { 1, 3, 6, 1, 2, 1, 25, 2, 3, 1, 4, 0 };
	oid mem_buffer_size[]      = { 1, 3, 6, 1, 2, 1, 25, 2, 3, 1, 6, 6 };
	oid mem_cache_size[]       = { 1, 3, 6, 1, 2, 1, 25, 2, 3, 1, 6, 7 };
	oid system_operation_oid[] = { 1, 3, 6, 1, 2, 1, 1, 1, 0 };

	op_get_memcache_servers();
	parse_db_conf();

	init_snmp("op-snmp-storage-stats");

	params = (struct snmp_check_parameters *)malloc(sizeof(struct snmp_check_parameters));
	if (params == NULL) {
		printf("Unable to allocate memory to parse command line parameters\n");
		return UNKNOWN;
	}

	status = op_snmp_parse_args(argc,argv,params);
	if (status != OK) {
		return usage();
	}

	snmp_sess_init( &session );
	status = prepare_snmp_session( &session, params, NULL );
	if (status != OK) {
		printf("Unable to locate community id %d\n",params->community_id);
		return UNKNOWN;
	}

	root_leaf = op_snmp_walk(disks_names_base_oid, sizeof(disks_names_base_oid), &session);
	if (root_leaf == NULL) {
		if (op_snmp_error != NULL) {
			printf("%s\n",op_snmp_error);
		} else {
			printf("Unable to retrieve object oid through snmp\n");
		}
		return UNKNOWN;
	}

	for(leaf = root_leaf; leaf; leaf = leaf->next) {
		if (strcasecmp(leaf->_char,params->filter) == 0) {
			index = leaf->index;
			break;
		}
	}

	if (index == 0) {

		// try to locate disks that starts with params->filter
		asprintf(&disk_regex_string,"^%s.*$",params->filter);
		if (regcomp(&disk_name_regex,disk_regex_string,0)) {
			printf("Invalid regular expression supplied %s\n",disk_regex_string);
			SOCK_CLEANUP;
			return UNKNOWN;
		}

		for(leaf = root_leaf; leaf; leaf = leaf->next) {
			if (!regexec(&disk_name_regex, leaf->_char, 0, NULL, 0)) {
				index = leaf->index;
				params->filter = leaf->_char;
				break;
			}
		}

		if (index == 0) {
			printf("Unable to locate %s counters through snmp\n",params->filter);
			SOCK_CLEANUP;
			return UNKNOWN;
		}
	}

	disk_used_size[11] = index;
	disk_max_size[11]  = index;
	disk_size_unit[11] = index;

	root_leaf  = op_snmp_get(disk_used_size, sizeof(disk_used_size), &session);
	if (root_leaf == NULL) {
		if (op_snmp_error != NULL) {
			printf("%s\n",op_snmp_error);
		} else {
			printf("Unable to retrieve object oid through snmp\n");
		}
		return UNKNOWN;
	}

	disk_usage = (float)root_leaf->_integer;

    root_leaf  = op_snmp_get(system_operation_oid, sizeof(system_operation_oid), &session);
    if (root_leaf == NULL) {
        if (op_snmp_error != NULL) {
            printf("%s\n",op_snmp_error);
        } else {
            printf("Unable to retrieve object oid through snmp\n");
        }
        return UNKNOWN;
    }

    if (! regcomp(&system_operation_regex, "Linux", 0)) {
        if (regexec(&system_operation_regex, root_leaf->_char, 0, NULL, 0) == 0) {
            isLinux = TRUE;
        }
    }

    if (isLinux) {
		if (regcomp(&disk_name_regex,"Physical memory",0)) {
			printf("Invalid regular expression supplied %s\n","memory");
			SOCK_CLEANUP;
			return UNKNOWN;
		}

        if (regexec(&disk_name_regex, params->filter, 0, NULL, 0) == 0) {
            root_leaf = op_snmp_get(mem_cache_size, sizeof(mem_cache_size), &session);
            if (root_leaf == NULL) {
    	        if (op_snmp_error != NULL) {
    		        printf("%s\n",op_snmp_error);
    	        } else {
        		    printf("Unable to retrieve object oid through snmp\n");
        	    }
        	    return UNKNOWN;
            }

            mem_cache = (float)root_leaf->_integer;
            root_leaf = op_snmp_get(mem_buffer_size, sizeof(mem_buffer_size), &session);
            if (root_leaf == NULL) {
    	        if (op_snmp_error != NULL) {
    		        printf("%s\n",op_snmp_error);
    	        } else {
        		    printf("Unable to retrieve object oid through snmp\n");
        	    }
        	    return UNKNOWN;
            }

            mem_buffer = (float)root_leaf->_integer;
            disk_usage -= (mem_cache + mem_buffer);
        }
    }

   	root_leaf  = op_snmp_get(disk_max_size, sizeof(disk_max_size), &session);
	if (root_leaf == NULL) {
		if (op_snmp_error != NULL) {
			printf("%s\n",op_snmp_error);
		} else {
			printf("Unable to retrieve object oid through snmp\n");
		}
		return UNKNOWN;
	}
	disk_size = (float)root_leaf->_integer;

	if (disk_size == 0) {
		printf("%s of zero size.\n",params->filter);
		return UNKNOWN;
	}

	root_leaf  = op_snmp_get(disk_size_unit, sizeof(disk_size_unit), &session);
	if (root_leaf == NULL) {
		if (op_snmp_error != NULL) {
			printf("%s\n",op_snmp_error);
		} else {
			printf("Unable to retrieve object oid through snmp\n");
		}
		return UNKNOWN;
	}
	disk_unit = (float)root_leaf->_integer;

	usage_pct = (disk_usage * 100) / disk_size;
	status = OK;
	if (usage_pct > params->critical)
		status = CRITICAL;
	else if (usage_pct > params->warning)
		status = WARNING;

	snprintf(plugin_output,
	    MAX_PLUGIN_OUTPUT_LENGTH - 1,
	    "%s disk usage %.2f %% (%s) |usage=%.2f%%;%.2f;%.2f;0;100\n",
	    params->filter,
	    usage_pct,
	    prepare_disk_size_output(disk_usage, disk_size, disk_unit),
	    usage_pct,
	    params->warning,
	    params->critical
	);

	printf("%s",plugin_output);

	return status;

}
