#include <time.h>
#include <stdlib.h>

#include <algorithm>
#include <iostream>
#include <vector>
#include <string>
#include <sstream>
#include <iomanip>
#include <map>

#include <opservices.h>

using namespace std;

#define PLUGIN_MAX_METRICS		40
#define PLUGIN_METRIC_FIELDS	8
#define PLUGIN_METRIC_NAME		1
#define PLUGIN_METRIC_VALUE		2
#define PLUGIN_METRIC_UNIT		3
#define PLUGIN_METRIC_WARNING	4
#define PLUGIN_METRIC_CRITICAL	5
#define PLUGIN_METRIC_MIN		6
#define PLUGIN_METRIC_MAX		7

void alarm_handler(int sig);
int run_check(char *command_line, char **output);
char* extract_check_result(FILE *fp);
int real_exit_code(int code);
char* escape_backslashes(char *rawbuf);
int parse_perfdata(char *perfdata, vector<Service_Metric> &metrics);
int get_last_service_metric_perfdata(Service_Metric base_metric,
									const Baseline &baseline,
									size_t check_interval,
									vector<Service_Metric> &perfdata);
int get_baseline_plugin_output(vector<Service_Metric> plugin_metrics,
								vector<Service_Metric> last_metrics,
								size_t check_interval,
								char **out_perfdata,
								vector<string> &baseline_msgs);
void get_baseline_broken_msg(double hw_value,
							double std_dev,
							Baseline baseline,
							Service_Metric metric,
							vector<string> &msgs);
int has_enough_hw(vector<Service_Metric> metrics, long need);
bool getOptions(int argc, char **argv);

map<string, string> args;
int verbose = 0;
string plugin_name;

void
usage() {
	cout << "./"
		<< plugin_name
		<< " -A <prediction_algorithm>"
		<< " -H <host_id>"
		<< " -s <service_id>"
		<< " -t <timeout>"
		<< " -i <check_interval>"
		<< " -C <command_line>"
		<< " [-v <0-4>]"
		<< endl;
}

void
help()
{
	cout
		<< endl
		<< "    Este plugin executa outro plugin, realiza o calculo do baseline para todas" << endl
		<< "as metricas retornadas que possuam baseline associado no OpMon." << endl
		<< endl
		<< "Uso:" << endl;
	usage();
	cout << endl;
	cout
		<< "  -A      Algoritmo utilizado para calcular o baseline. As opcoes disponiveis" << endl
		<< "        sao: holts e holtwinters (padrao)." << endl
		<< "  -C      Linha de comando com o plugin que sera executado." << endl
		<< "  -H      ID do host onde o resultado da checagem sera enviado." << endl
		<< "  -s      ID do servico para o qual o resultado da checagem deve ser enviado." << endl
		<< "  -h      Exibe esta ajuda." << endl
		<< "  -i      Intervalo de checagem configurado no OpMon, frequencia de execucao do" << endl
		<< "        plugin. Utilizado para gerar os intervalos de tempo utilizados para" << endl
		<< "        selecionar as amostras para o calculo do baseline." << endl
		<< "  -t      Timeout para a execucao do plugin em segundos." << endl
		<< "  -v      Nivel de verbose utilizado para exibir mensagens de debug." << endl
	   	<< endl;
}

bool getOptions(int argc, char **argv)
{
	args["algorithm"].assign("holtwinters");

	for (int i=1; i < argc; i++) {
		if (strcmp(argv[i], "-H") == 0) {
			if ((i+1) >= argc) {
				return false;
			}

			args["host_id"].assign( argv[i+1] );
		} else if (strcmp(argv[i], "-A") == 0) {
			if ((i+1) >= argc) {
				return false;
			}

			args["algorithm"].assign( argv[i+1] );
			transform(args["algorithm"].begin(),
				args["algorithm"].end(),
				args["algorithm"].begin(),
				::tolower
			);
		} else if (strcmp(argv[i], "-s") == 0) {
			if ((i+1) >= argc) {
				return false;
			}

			args["service_id"].assign( argv[i+1] );
		} else if (strcmp(argv[i], "-t") == 0) {
			if ((i+1) >= argc) {
				return false;
			}

			args["timeout"].assign( argv[i+1] );
		} else if (strcmp(argv[i], "-i") == 0) {
			if ((i+1) >= argc) {
				return false;
			}

			args["check_interval"].assign( argv[i+1] );
		} else if (strcmp(argv[i], "-C") == 0) {
			if ((i+1) >= argc) {
				return false;
			}

			args["command_line"].assign( argv[i+1] );
		} else if (strcmp(argv[i], "-v") == 0) {
			if ((i+1) >= argc) {
				return false;
			}

			verbose = atoi(argv[i+1]);
		} else if (strcmp(argv[i], "-h") == 0) {
			help();
			exit(0);
		}
	}

	if ((args["algorithm"] != "holts")
		&& (args["algorithm"] != "holtwinters")) {
		return false;
	}

	return ((args["host_id"].empty())
			|| (args["service_id"].empty())
			|| (args["timeout"].empty())
			|| (args["check_interval"].empty())) ? false : true;
}

int
main(int argc, char *argv[])
{

	size_t	host_id;
	size_t	service_id;
	size_t	timeout;
	size_t	check_interval;
	size_t	i, j;
	int		return_code;
	int		number_svc_metrics;
	char	*plugin_output;
	char	*command_line;
	char	*perfdata;
	string	baseline_output;
	Service_Metric			**service_metrics = NULL;
	vector<Service_Metric>	plugin_metrics;
	vector<Service_Metric>	last_metrics;
	vector<string>			baseline_msgs;

	plugin_name = basename(argv[0]);

	/* parse database conf */
	return_code = parse_db_conf();
	if (return_code == ERROR) {
		cout << "Unable to parse database configuration file" << endl;
		exit(UNKNOWN);
	}

	if ( ! getOptions(argc, argv)) {
		usage();
		exit(UNKNOWN);
	}

	host_id			= atoi(args["host_id"].c_str());
	service_id		= atoi(args["service_id"].c_str());
	timeout			= atoi(args["timeout"].c_str());
	check_interval	= atoi(args["check_interval"].c_str());
	command_line	= strdup(args["command_line"].c_str());

	number_svc_metrics = count_service_metrics(host_id, service_id);
	if (number_svc_metrics <= 0) {
		cout << "This service doesn't have metrics in OpMon's database." << endl;
		return 1;
	}

	signal(SIGALRM, alarm_handler);
	alarm(timeout);
	return_code = run_check(command_line, &plugin_output);

	if (plugin_output == NULL) {
		cout << "Empty plugin return" << endl;
		exit(UNKNOWN);
	}

	strtok(plugin_output,"|");
	perfdata = strtok(NULL, "|");

	if (perfdata == NULL) {
		cout << plugin_output << endl;
		exit(return_code);
	}

	if (parse_perfdata(perfdata, plugin_metrics) != OK) {
		cout << plugin_output << endl;
		exit(return_code);
	}

	service_metrics = (Service_Metric**)calloc(number_svc_metrics, sizeof(Service_Metric*));
	for (i=0; (int)i < number_svc_metrics; i++) {
		service_metrics[i] = (Service_Metric*)calloc(1, sizeof(Service_Metric));
	}

	get_metrics_by_host_and_service_id(host_id, service_id, service_metrics);

	for (i=0; i < plugin_metrics.size(); i++) {
		for (j=0; (int)j < number_svc_metrics; j++) {
			return_code = strcmp(
				service_metrics[j]->metric_name,
				plugin_metrics.at(i).metric_name
			);

			if (return_code != 0) {
				continue;
			}

			plugin_metrics.at(i).host_id		= service_metrics[j]->host_id;
			plugin_metrics.at(i).service_id		= service_metrics[j]->service_id;
			plugin_metrics.at(i).metric_id		= service_metrics[j]->metric_id;
			plugin_metrics.at(i).baseline_id	= service_metrics[j]->baseline_id;

			break;
		}
	}

	for (i=0; (int)i < number_svc_metrics; i++) {
		free(service_metrics[i]);
	}
	free(service_metrics);

	return_code = get_baseline_plugin_output(
		plugin_metrics,
		last_metrics,
		check_interval * 60,
		&perfdata,
		baseline_msgs
	);

	if (return_code > 0) {
		baseline_output.assign("Baseline broken: ");

		for (i=0; i < baseline_msgs.size(); i++) {
			baseline_output.append(baseline_msgs.at(i) + " ");
		}

		baseline_output.append("- ");
	}
	cout << baseline_output << strtok(plugin_output,"|") << "|" << perfdata << endl;

	return return_code;
}

int
has_hw_state(vector<Service_Metric> metrics)
{

	size_t	i;
	int		null_state;

	for (i=0; i < metrics.size(); i++) {

		null_state = (metrics.at(i).baseline != 0);
		null_state = ((null_state) && (metrics.at(i).lower_limit != 0));
		null_state = ((null_state) && (metrics.at(i).upper_limit != 0));

		if (null_state) {
			return 1;
		}
	}

	return 0;
}

int
get_baseline_plugin_output(vector<Service_Metric> plugin_metrics,
							vector<Service_Metric> last_metrics,
							size_t check_interval,
							char **out_perfdata,
							vector<string> &baseline_msgs)
{
	size_t		i, j;
	size_t		size;
	int			return_code;
	int			check_exit_code	= OK;
	int			exit_code_tmp	= OK;
	double		*perfdata;
	Baseline	baseline;
	double		hw_value = 0;
	double		std_dev;
	char		*msg	= NULL;

	(*out_perfdata) = NULL;

	for (i=0; i < plugin_metrics.size(); i++) {
		asprintf(
			out_perfdata,
			"%s %s=%s%s;%s;%s;%s;%s",
			((*out_perfdata) != NULL) ? (*out_perfdata) : "",
			plugin_metrics.at(i).metric_name,
			plugin_metrics.at(i).value,
			plugin_metrics.at(i).unit,
			plugin_metrics.at(i).warning,
			plugin_metrics.at(i).critical,
			plugin_metrics.at(i).min,
			plugin_metrics.at(i).max
		);

		if ((plugin_metrics.at(i).baseline_id == 0)
			|| (get_baseline(plugin_metrics.at(i).baseline_id, &baseline) != OK)) {
			continue;
		}

		return_code = get_last_service_metric_perfdata(
			plugin_metrics.at(i),
			baseline,
			check_interval,
			last_metrics
		);

		if (return_code == ERROR) {
			continue;
		}

		size = last_metrics.size();
		perfdata = (double*)calloc(size, sizeof(double));

		for (j=0; j < last_metrics.size(); j++) {
			perfdata[j] = atof(last_metrics.at(j).value);
		}

		if (args["algorithm"] == "holtwinters") {
			hw_value = holtwinters(perfdata, size);
		} else if (args["algorithm"] == "holts") {
			hw_value = holts(perfdata, size);
		}

		std_dev		= standard_deviation(perfdata, (int)size);

		exit_code_tmp = get_baseline_service_status(
			hw_value,
			std_dev,
			baseline,
			atof(plugin_metrics.at(i).value),
			plugin_metrics.at(i).metric_name,
			&msg
		);

		if (msg != NULL) {
			baseline_msgs.push_back(string(msg));
			free(msg);
			msg = NULL;
		}

		if ((check_exit_code == OK) && (exit_code_tmp == CRITICAL)) {
			check_exit_code = CRITICAL;
		}
		free(perfdata);

		asprintf(
			out_perfdata,
			"%s;%.2f;%.2f;%.2f",
			(*out_perfdata),
			hw_value,
			hw_value - std_dev * baseline.multiply_down,
			hw_value + std_dev * baseline.multiply_up
		);
	}

	return check_exit_code;
}

/* run a check */
int
run_check(char *command_line, char **output)
{

	FILE	*fp;
	int		exit_code;

	fp = popen(command_line, "r");
	if (fp == NULL) {
		fprintf(stdout, "Unable to run specified command line.\n");
		exit(UNKNOWN);
	}

	/* extract check result */
	*output = extract_check_result(fp);

	/* close the process */
	exit_code = pclose(fp);

	return real_exit_code(exit_code);

}

/* extract check result */
char*
extract_check_result(FILE *fp)
{

	int 	size;
	char	buffer[OP_PLUGIN_BUFFERSIZE] = "";
	char	output[OP_PLUGIN_BUFFERSIZE] = "";

	/* get all lines of plugin output - escape newlines */
	size = OP_MAX_OUTPUT;

	while((size > 0) && (fgets(buffer,sizeof(buffer)-1,fp))) {
		strncat(output, buffer, size);
		size -= strlen(buffer);
	}

	return escape_backslashes(output);
}

/* escapes newlines (\n), tabs (\t) and carriage return (\r) in a string */
char*
escape_backslashes(char *rawbuf)
{

	char			*newbuf = NULL;
	register int	x,y;

	if (rawbuf == NULL) {
		return NULL;
	}

	/* allocate enough memory to escape all chars if necessary */
	if ((newbuf = (char*)malloc((strlen(rawbuf)*2)+1)) == NULL) {
		return NULL;
	}

	for(x=0, y=0; rawbuf[x] != (char)'\x0'; x++){

		/* escape backslashes */
		if (rawbuf[x]=='\\') {
			newbuf[y++]='\\';
			newbuf[y++]='\\';

		} else if (rawbuf[x]=='\n') {
			newbuf[y++]=' ';
		} else {
			newbuf[y++]=rawbuf[x];
		}
	}

	newbuf[y] = '\x0';

	return newbuf;
}

/* called when check runs into timeout */
void
alarm_handler(int sig)
{
	printf("(Service Check Timed)\n");
	exit(UNKNOWN);
}

/* convert exit code to int */
int
real_exit_code(int code)
{

	if (code == -1) {
		return UNKNOWN;
	}

	if (WIFSIGNALED(code)) {
		return(128 + WTERMSIG(code));
	}

	if (WIFEXITED(code)) {
		return(WEXITSTATUS(code));
	}

	return OK;
}

int
parse_perfdata(char *perfdata, vector<Service_Metric> &metrics)
{

	int				ret, i, length;
	char			*metric_string	= NULL;
	char			*aux_string		= NULL;
	Service_Metric	*metric;
	regex_t			regex;
	regex_t			metric_regex;
	regmatch_t		pmatch[PLUGIN_MAX_METRICS]; // Maximum number of metrics

	// regex to split by metric
	// OK: \PhysicalDisk(0 C:)\Disk Transfers/sec: 0|'\PhysicalDisk(0 C:)\Disk Transfers/sec'=0;20;25;
	ret = regcomp(&regex," *([^=]+=[^ ]+)", REG_EXTENDED); 
	if (ret != 0) {
		cout << "Unable to compile regular expression." << endl;
		return ERROR;
	}

	// a regex to extract our metric values 
	ret = regcomp(
		&metric_regex,
		"([^=]+)=([0-9.]+)([^;]*;?)?([^;]*;?)?([^;]*;?)?([^;]*;?)?([^;]*)",
		REG_EXTENDED
	);

	if (ret != 0) {
		cout << "Unable to compile internal regular expression." << endl;
		return ERROR;
	}

	// for every metric, we gonna extract our data
	while(regexec(&regex, perfdata, PLUGIN_MAX_METRICS, pmatch, 0) == 0 ) {

		length = pmatch[1].rm_eo - pmatch[1].rm_so + 1;

		// create a string to store our metric data
		metric_string = (char*)calloc(length,sizeof(char));

		// copy metric data to string
		strncpy(
			metric_string,
			perfdata + pmatch[1].rm_so,
			pmatch[1].rm_eo - pmatch[1].rm_so
		);

		// advance perfdata to the next metric
		perfdata += pmatch[1].rm_eo;

		// apply the regex to one metric data
		if (regexec(&metric_regex, metric_string, PLUGIN_MAX_METRICS, pmatch, 0) != 0) {
			free(metric_string);
			continue;
		}

		metric = (Service_Metric*)calloc(1, sizeof(Service_Metric));

		// iterate through our matches
		for(i=1; i <= PLUGIN_METRIC_FIELDS; i++) {

			// end of matches
			if (pmatch[i].rm_so < 0) {
				break;
			}

			length = pmatch[i].rm_eo - pmatch[i].rm_so + 1;

			aux_string = (char*)calloc(length, sizeof(char));
			strncpy(
				aux_string,
				metric_string + pmatch[i].rm_so,
				pmatch[i].rm_eo - pmatch[i].rm_so
			);

			remove_invalid_char(&aux_string);
			switch(i) {
				case PLUGIN_METRIC_NAME:
					metric->metric_name = strdup(aux_string);
					break;

				case PLUGIN_METRIC_VALUE:
					metric->value = strdup(aux_string);
					break;

				case PLUGIN_METRIC_UNIT:
					metric->unit = strdup(aux_string);
					break;

				case PLUGIN_METRIC_WARNING:
					metric->warning = strdup(aux_string);
					break;

				case PLUGIN_METRIC_CRITICAL:
					metric->critical = strdup(aux_string);
					break;

				case PLUGIN_METRIC_MIN:
					metric->min = strdup(aux_string);
					break;

				case PLUGIN_METRIC_MAX:
					metric->max = strdup(aux_string);
					break;
			}

			free(aux_string);
		}

		metrics.push_back((*metric));

		free(metric);
		free(metric_string);
	}

	regfree(&metric_regex);
	regfree(&regex);

	return OK;

}

void
freeBaselineCalendar(int ***calendar, int size)
{
	for (int i=0; i < size; i++) {
		free((*calendar)[i]);
	}

	free((*calendar));
}

int
get_last_service_metric_perfdata(Service_Metric base_metric,
								const Baseline &baseline,
								size_t check_interval,
								vector<Service_Metric> &perfdata)
{
	char			*query;
	char			*gaps = NULL;
	char			db_opperf[] = "opperf";
	int				i;
//	int				j;
	time_t			check_timestamp;
	int				num_fields;
	int				return_length;
	int				**calendar;
	MYSQL_RES		*res = NULL;
	MYSQL_ROW		row;
	Service_Metric	*perfdata_tmp;
	string			last_date;
	int				metrics_count=0;
	int				found_samples=0;


	if ((base_metric.host_id <= 0) || (base_metric.service_id <= 0)
			|| (base_metric.metric_id <= 0) || (baseline.samples <= 0)) {
		return ERROR;
	}

	time(&check_timestamp);

	if (baseline.type == 3)	 {
		calendar = gen_baseline_calendar(baseline, check_timestamp);
		create_baseline_calendar_time_gaps(check_timestamp, check_interval, baseline, &calendar, &gaps);
/*
		for (i=0; i < baseline.samples; i++) {
			for (j=0; j < 31; j++) {
				printf("%02d ",calendar[i][j]);
			}
			cout << endl;
		}
*/
		freeBaselineCalendar(&calendar, baseline.samples * 2);
	} else {
		create_baseline_time_gaps(check_timestamp, check_interval, baseline, &gaps);
	}

	if (gaps == NULL) {
		return ERROR;
	}

	asprintf(
		&query,
		"SELECT *"
		" FROM service_perf_%lu"
		" WHERE service_id=%lu"
		"   AND metric_id=%lu"
		"   AND (%s)"
		" ORDER BY entry_time DESC",
		base_metric.host_id,
		base_metric.service_id,
		base_metric.metric_id,
		gaps
	);

	parse_db_conf();

	if (verbose > 1) {
		cout << query << endl;
	}

	return_length = do_query(query, DBHOST, DBUSER, DBPASS, db_opperf, &res);

	if (return_length <= 0) {
		free(query);
		return ERROR;
	}

	free(query);

	num_fields = mysql_num_fields(res);

	while ((row = mysql_fetch_row(res))) {

		perfdata_tmp = (Service_Metric*) calloc(return_length, sizeof(Service_Metric));
		perfdata_tmp->service_id = base_metric.host_id;

		for (i=0; i < num_fields; i++) {

			if ((row[i] == NULL) || (strlen(row[i]) == 0)) {
				continue;
			}

			switch (i) {
				case FIELD_SERVICE_PERF_ENTRY_TIME:
					if (verbose > 0) {
						if (strcmp(strtok(row[i], " "), last_date.c_str()) != 0) {
							if ( ! last_date.empty()) {
								cout << " Total: " << metrics_count;
								found_samples += metrics_count;
							}
							cout << endl << "[" << row[i] << "] ";
							last_date.assign(strtok(row[i], " "));
							metrics_count = 0;
						}

					}
					continue;
					break;
				case FIELD_SERVICE_PERF_SERVICE_ID:
					perfdata_tmp->service_id = atoi(row[i]);
					break;
				case FIELD_SERVICE_PERF_METRIC_ID:
					perfdata_tmp->metric_id = atoi(row[i]);
					break;
				case FIELD_SERVICE_PERF_PERF_VALUE:
					perfdata_tmp->value = strdup(row[i]);
					if (verbose > 0) {
						cout << row[i] << " ";
						metrics_count++;
					}
					break;
				case FIELD_SERVICE_PERF_WARNING:
					perfdata_tmp->warning = strdup(row[i]);
					break;
				case FIELD_SERVICE_PERF_CRITICAL:
					perfdata_tmp->critical = strdup(row[i]);
					break;
				case FIELD_SERVICE_PERF_BASELINE:
					perfdata_tmp->baseline = atof(row[i]);
					break;
				case FIELD_SERVICE_PERF_LOWER_LIMIT:
					perfdata_tmp->lower_limit = atof(row[i]);
					break;
				case FIELD_SERVICE_PERF_UPPER_LIMIT:
					perfdata_tmp->upper_limit = atof(row[i]);
					break;
			}

		}

		perfdata.push_back((*perfdata_tmp));
		free(perfdata_tmp);

	}

	mysql_free_result(res);

	reverse(perfdata.begin(), perfdata.end());

	if ((verbose > 0) && ( ! last_date.empty())) {
		cout << " Total: " << metrics_count;
		found_samples += metrics_count;
		cout << endl << "Found samples: " << found_samples << endl;
	}

	return OK;
}

